﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file PlotVertices2.hpp
 * @brief генератор вершин однопараметрического графика на плоскости
 */
#pragma once
#include <avo/Vec2.hpp>
#include <vector>
#include <algorithm>
#include <limits>
#include <type_traits>
#include <cmath>
#include <cassert>

using Scalar   = double;
using Point_2  = avo::Vec_2<Scalar>;
using Points_2 = std::vector<Point_2>;

/// генератор вершин графика на плоскости
/**
 * @param fxy : R -> R2 непрерывная на [t0, t1] функция, возвращающая точку плоскости для некоторого значения параметра
 * @param t0 начальное значение параметра
 * @param t1 конечное значение параметра
 * @param min_t_step минимальный возможный шаг параметра
 * @param max_t_step максимальный возможный шаг параметра
 * @param tol (условная) допустимая относительная погрешность на звено (оценочная длина куска графика / длина отрезка)
 */
template <class FXY>
Points_2 plot_vertices2(FXY fxy, Scalar t0, Scalar t1, Scalar min_t_step, Scalar max_t_step, Scalar tol)
{
  using namespace std;
  static_assert (
    is_convertible<decltype(fxy(t0).x), Scalar>::value &&
    is_convertible<decltype(fxy(t0).y), Scalar>::value,
    "APlot2D plot_vertices2: fxy parameter must be a function taking Scalar values and returning values which have .x and .y defined"
  );

  assert(t0 <= t1);
  assert(min_t_step <= max_t_step);
  assert(0. < min_t_step);
  assert(isfinite(max_t_step));
  assert(0. <= tol);

  // yeah, magic numbers dwell here...
  static const auto
    ONE                 = Scalar(1.),
    SENSE_FRACTION      = Scalar(1./8.),
    SHRINK_FACTOR       = Scalar(0.25),
    RELAX_FACTOR        = Scalar(1.25),
    NORMAL_TOL_FACTOR   = Scalar(1./32.),
    MIN_STEP_EPS_FACTOR = Scalar(3.);


  min_t_step = fmax(min_t_step, MIN_STEP_EPS_FACTOR * numeric_limits<Scalar>::epsilon() * (abs(t0) + abs(t1)));
  // "нормальный" шаг выбирается как среднее геометрическое минимального и максимального шагов
  const auto normal_t_step = sqrt(min_t_step) * sqrt(max_t_step),
             normal_tol = NORMAL_TOL_FACTOR * tol;

  auto t = t0,
       t_step = normal_t_step;
  auto p = fxy(t);

  Points_2 ps;
  ps.emplace_back(p.x, p.y);
  while (t < t1)
  {
    const auto next_t = fmin(t1, t + t_step);
    const auto next_p = fxy(next_t);

    bool add_next = t_step == min_t_step;
    if (!add_next)
    {
      const auto dl = hypot(p.x - next_p.x, p.y - next_p.y);

      const auto sense_t = (ONE - SENSE_FRACTION) * t + SENSE_FRACTION * next_t;
      auto s = fxy(sense_t);
      
      auto sl = hypot(p.x - s.x, p.y - s.y) - dl; // !заранее вычли dl из sl
      for (int i = 2; i < 8; ++i)
      {
        const auto iF = SENSE_FRACTION * i;
        const auto sense_t = iF * next_t + (ONE - iF) * t;
        const auto s2 = fxy(sense_t);
        sl += hypot(s.x - s2.x, s.y - s2.y);
        s = s2;
      }
      sl += hypot(s.x - next_p.x, s.y - next_p.y);

      add_next = dl * tol >= sl;
      if (!add_next)
        t_step = fmax(min_t_step, SHRINK_FACTOR * t_step);
      else if (sl < dl * normal_tol)
        t_step = fmin(max_t_step, RELAX_FACTOR * t_step);
    }
    else // уже достигнут минимальный шаг, попробуем его немного отпустить
      t_step = fmin(max_t_step, RELAX_FACTOR * t_step);

    if (add_next)
    {
      t = next_t;
      p = next_p;
      ps.emplace_back(next_p.x, next_p.y);
    }
  }

  // очистить от NaN'ов
  ps.erase(remove_if(ps.begin(), ps.end(),
    [](const Point_2 &p) { return isnan(p.x) || isnan(p.y); }), ps.end());
  return ps;
}

/// генератор вершин графика на плоскости (разъединённые функции координат)
/**
 * @param fx : R -> R непрерывная на [t0, t1] функция, возвращающая x-координату для некоторого значения параметра
 * @param fy : R -> R непрерывная на [t0, t1] функция, возвращающая y-координату для некоторого значения параметра
 * @param t0 начальное значение параметра
 * @param t1 конечное значение параметра
 * @param min_t_step минимальный возможный шаг параметра
 * @param max_t_step максимальный возможный шаг параметра
 * @param tol (условная) допустимая относительная погрешность на звено (оценочная длина куска графика / длина отрезка)
 */
template <class FX, class FY>
inline Points_2 plot_vertices2(FX fx, FY fy, Scalar t0, Scalar t1, Scalar min_t_step, Scalar max_t_step, Scalar tol)
{
  return plot_vertices2([&fx, &fy](Scalar t) { return Point_2(fx(t), fy(t)); },
    t0, t1, min_t_step, max_t_step, tol);
}
