﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: AStarVis
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Визуализатор алгоритма поиска A*.
 * Поиск пути производится на графе, построенном для некоторой псевдослучайной карты высот.
 * Также см. https://evetro.wordpress.com/2011/05/15/av-2/
 * Нажатие
 *  - SPACE/ENTER генерирует новый ландшафт и вычисляет пути;
 *  - ESC выход;
 *  - стрелки, WASDXZ, IJKLOP -- движение камеры;
 *  - 0, 1, 2, ..., 9 выбор множителя эвристики А*;
 *  - G -- показывать граф;
 *  - P -- показывать пути, вычисленные алгоритмами Дейкстры, А*.
 */
#include <GL/freeglut.h>
#include <iostream>

#include <avo/SphericalCamera.hpp>
#include <avo/Timer.hpp>

#include "Scene.hpp" // сцена AStarVis


//
// глобальные переменные
//

/// последний зафиксированный момент времени
double t;
/// задержка между визуализацией событий (секунды)
/**
 * Необходимо помнить, что частота кадров 1/timedelay в секунду -- недостижимый идеал,
 * так как время на отрисовку не учитывается в этой задержке.
 */
double timedelay = 0.01;

/// камера (задаётся двумя сферическими углами и расстоянием до начала координат)
avo::Spherical_camera camera(avo::PI_div_2 / 2., avo::PI_div_2 / 3., 600.);
/// сцена AStarVis
Scene scene;


//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  scene.paint();
  glFinish();
  glutSwapBuffers();
}


/// обработка изменений размеров окна
void reshape(int new_width, int new_height)
{
  // новая матрица проекции
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();   
  glLoadIdentity();
  // "магические" константы: 75 градусов -- угол зрения камеры,
  // 20 и 4116 -- расстояния до ближней и дальней отсекающих плоскостей,
  // таким образом, видно объекты на расстояниях [20, 4116] -- 4096 единиц глубины
  gluPerspective(75, double(new_width) / double(new_height), 20.0, 4116.0);
  glPushMatrix();
  camera.apply();
  // область просмотра == клиентская область окна
  glViewport(0, 0, new_width, new_height);
}


/// вычислить состояние сцены на следующий кадр
void frame_move()
{
  const double t1 = avo::seconds();
  if (t1 - t >= timedelay)
  {
    int i = static_cast<int>(1.0 + std::min(1e+6, std::floor((t1 - t) / timedelay)));
    while (i --> 0 && scene.next_event()) {}
    t = t1;
    glutPostRedisplay(); // уведомить FreeGlut, что пора перерисовать сцену
  }
}


/// обработать нажатие "обычной" клавиши
void on_key(unsigned char key, int x, int y)
{
  switch (key)
  {
  case 'w': case 'W': // вперёд
    camera.forward();
    camera.apply();
    break;
  case 's': case 'S': // назад
    camera.backward();
    camera.apply();
    break;
  case 'a': case 'A': // повернуть, двигая камеру влево
    camera.left();
    camera.apply();
    break;
  case 'd': case 'D': // повернуть, двигая камеру вправо
    camera.right();
    camera.apply();
    break;
  case 'z': case 'Z': // повернуть, двигая камеру вверх
    camera.up();
    camera.apply();
    break;
  case 'x': case 'X': // повернуть, двигая камеру вниз
    camera.down();
    camera.apply();
    break;
  
  case 'i': case 'I': // передвинуть начало координат вперёд
    camera.origin_forward(50.);
    camera.apply();
    break;
  case 'k': case 'K': // передвинуть начало координат назад
    camera.origin_forward(-50.);
    camera.apply();
    break;
  case 'j': case 'J': // передвинуть начало координат влево
    camera.origin_right(-30.);
    camera.apply();
    break;
  case 'l': case 'L': // передвинуть начало координат вправо
    camera.origin_right(30.);
    camera.apply();
    break;
  case 'o': case 'O': // передвинуть начало координат вверх
    camera.origin_up(30.);
    camera.apply();
    break;
  case 'p': case 'P': // передвинуть начало координат вниз
    camera.origin_up(-30.);
    camera.apply();
    break;

  case '=':
  case '+': // увеличить скорость визуализации
    timedelay *= 0.5;
    if (timedelay < 1e-6)
      timedelay = 1.0 / 65356;
    break;

  case '-': // уменьшить скорость визуализации
    timedelay *= 2.0; // может достигнуть бесконечности (не проверяется)
    break;

  case 'g': case 'G': // включить или выключить отрисовку рёбер графа
    scene.paintG = !scene.paintG;
    break;

  case 'u': case 'U': // включить или выключить отрисовку путей
    scene.paintApath = scene.paintDpath = !scene.paintApath;
    break;

  case '1': case '2': case '3': case '4': case '5':
  case '6': case '7': case '8': case '9': case '0':
    // выбрать множитель эвристики А* (соответствующее целое число)
    scene.change_factor(static_cast<float>(key - '0'));
    break;

  case 'b': case 'B': // переключить двунаправленный/однонаправленный А*
    scene.switch_algorithm();
    break;

  case 27: // ESC ASCII-код, покинуть программу
    glutLeaveMainLoop();
    break;

  case ' ': case '\r': case '\n': // сгенерировать новый ландшафт и построить пути
    scene.new_terrain_and_paths();
    break;
  
  default:
    return;
  }
}


/// обработать нажатие "специальной" клавиши
void on_special_key(int key, int x, int y)
{
  switch (key)
  {
  case GLUT_KEY_LEFT:
    camera.left();
    camera.apply();
    break;

  case GLUT_KEY_UP:
    camera.up();
    camera.apply();
    break;

  case GLUT_KEY_RIGHT:
    camera.right();
    camera.apply();
    break;

  case GLUT_KEY_DOWN:
    camera.down();
    camera.apply();
    break;

  default:
    return;
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);

  // создать окно с привязанным контекстом OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
  glutInitWindowPosition(10, 10);
  glutInitWindowSize(1000, 800);
  glutCreateWindow("A* visualization");

  // параметры OpenGL для данного примера
  glShadeModel(GL_FLAT);
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  
  glEnable(GL_DEPTH_TEST); // использовать z-buffer
  glEnable(GL_LIGHTING);   // вершинное освещение

  // включить рисование примитивов, заданных массивами
  glEnableClientState(GL_COLOR_ARRAY);  // включить массив цветов вершин
  glEnableClientState(GL_NORMAL_ARRAY); // включить массив нормалей при вершинах
  glEnableClientState(GL_VERTEX_ARRAY); // включить массив координат вершин

  // источник света №0
  glEnable(GL_LIGHT0);
  float pos0[] = { 0.25f, 0.5f, 1.0f, 0.0f };
  glLightfv(GL_LIGHT0, GL_POSITION, pos0);
  float ambCol[] = { 0.1f, 0.1f, 0.1f, 1.0f };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambCol);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

  // "центрирование" камеры
  glMatrixMode(GL_MODELVIEW);
  glTranslatef(-0.5f * scene.x_size(), -0.5f * scene.y_size(), 0.0f);

  // привязка функций-обработчиков событий
  glutDisplayFunc(repaint);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(on_key);
  glutSpecialFunc(on_special_key);
  glutIdleFunc(frame_move);

  // старт
  scene.new_terrain_and_paths();

  t = avo::seconds();
  glutMainLoop();
  std::cout << "\nBye!";
  return 0;
}
