﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: GLPrimitivesVA
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Рисует случайно сгенерированный движущийся набор цветных точек ("вершин"), соединённых
 * в GL-примитив некоторого вида (выбирается цифровой клавишей):
 *   1 -- GL_POINTS
 *   2 -- GL_LINES
 *   3 -- GL_LINE_STRIP
 *   4 -- GL_LINE_LOOP
 *   5 -- GL_TRIANGLES
 *   6 -- GL_TRIANGLE_STRIP
 *   7 -- GL_TRIANGLE_FAN
 *   8 -- GL_QUADS
 *   9 -- GL_QUAD_STRIP
 *
 * Демонстрирует использование массивов вершинных данных.
 *
 * Вершины двигаются внутри коробки (0, 0, 0) -- (1, 1, 1), "отскакивая" от стенок.
 *
 * Нажатие BACKSPACE переключает режим закраски между FLAT и SMOOTH.
 * Нажатие +/- добавляет/удаляет вершины.
 *
 * Клавиши F10, F11 и F12 переключают режим растеризации полигонов (точки, линии, зарашенные).
 *
 * Нажатие ENTER/SPACE генерирует новый набор псевдослучайных вершин.
 * Нажатие ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>

#include <avo/Basics.hpp> // random
#include <avo/Timer.hpp>  // sleep, seconds
#include <avo/Vec3.hpp>   // 3D-вектор
#include <avo/Pixel.hpp>  // 32-битный RGBA-цвет

using avo::Vec_3f;
using avo::Pixel;
using avo::random;


/// сгенерировать некоторый случайный "яркий" цвет
inline Pixel random_bright_color()
{
  const avo::Pixel_component comp0 = random(0, 255),
                             comp1 = random(0, 255);
  switch (std::rand() % 3)
  {
  case 0: return Pixel::rgb(255, comp0, comp1);
  case 1: return Pixel::rgb(comp0, 255, comp1);
  case 2: return Pixel::rgb(comp0, comp1, 255);
  default: return Pixel::grey(255); // чтобы компилятор не писал warning'и
  }
}


/// максимально возможное число вершин
static const unsigned MAX_SIZE = 4096;

/// геометрический 3D-примитив с привязкой 32-битных цветов к каждой вершине (фиксированный максимальный размер)
struct V3f_C4ub
{
  GLenum mode;   ///< тип примитива
  unsigned size; ///< количество задействованных вершин

  Pixel  color[MAX_SIZE];    ///< цвета
  Vec_3f position[MAX_SIZE]; ///< позиции вершин

  /// Нарисовать.
  void paint() const
  {
    // Подключить массивы:
    // количество компонент, тип компоненты, расстояние между соседними элементами, указатель на первый элемент
    glColorPointer  (4, GL_UNSIGNED_BYTE, sizeof(Pixel), color);
    glVertexPointer (3, GL_FLOAT, sizeof(Vec_3f), position);

    // передать примитив: тип, с какого начинать, количество
    glDrawArrays(mode, 0, size);
  }
};


/// структура, управляющая движением частиц ("сцена")
struct Scene
{
  V3f_C4ub geometry;         ///< примитив
  Vec_3f velocity[MAX_SIZE]; ///< текущие скорости вершин

  /// получить текущий размер (количество активных вершин)
  unsigned size() const
  {
    return geometry.size;
  }

  /// установить новый размер
  void resize(unsigned new_size)
  {
    geometry.size = avo::clamp(new_size, 1u, MAX_SIZE); // не меньше одной, не больше MAX_SIZE
  }

  /// сгенерировать случайные вершины
  void make_random()
  {
    // цвета
    for (unsigned i = 0; i < MAX_SIZE; ++i)
      geometry.color[i] = random_bright_color();
    // позиции
    for (unsigned i = 0; i < MAX_SIZE; ++i)
      geometry.position[i] = Vec_3f(random(), random(), random());
    // скорости
    for (unsigned i = 0; i < MAX_SIZE; ++i)
      velocity[i] = Vec_3f(random(), random(), random()).normalize() * 0.05f;
  }

  /// пересчитать позиции с учётом пройденного времени (аргумент t_step)
  void move(float t_step)
  {
    for (unsigned i = 0, sz = size(); i < sz; ++i)
    {
      // собственно движение
      auto &veli = velocity[i];
      const auto next_position = (geometry.position[i] += t_step * veli);
      // отражение от стенок
      if      (next_position.x <= 0.f)  veli.x =  abs(veli.x);
      else if (1.f <= next_position.x)  veli.x = -abs(veli.x);
      if      (next_position.y <= 0.f)  veli.y =  abs(veli.y);
      else if (1.f <= next_position.y)  veli.y = -abs(veli.y);
      if      (next_position.z <= 0.f)  veli.z =  abs(veli.z);
      else if (1.f <= next_position.z)  veli.z = -abs(veli.z);
    }
  }

  /// нарисовать сцену (переадресация geometry)
  void paint() const
  {
    geometry.paint();
  }

  /// поменять тип примитива
  void mode(GLenum new_mode)
  {
    geometry.mode = new_mode;
  }
};


/// глобальный объект "сцена"
Scene scene;
/// момент отрисовки предыдущего кадра
double t0;
/// признак: используется градиентная заливка треугольников
bool smooth_shading;


// обработчики событий FreeGlut

/// изменение размеров окна, функция получает "новые размеры" (скопировано из GLPrimitives)
void reshape(int w, int h)
{
  // новая матрица проекции
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity(); // загрузить единичную матрицу
  /* угол зрения по вертикали, соотношение сторон экрана, минимальная дальность, максимальная дальность */
  gluPerspective(60, double(w) / double(h), 0.25, 2.0); // домножить на матрицу перспективной коррекции
  gluLookAt
  (
      0.5, 0.5, -0.625  // позиция камеры
    , 0.5, 0.5, 0.5     // точка, на которую смотрит камера
    , 0.0, 1.0, 0.0     // вектор "вверх"
  );

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  // установить координаты окна просмотра (вся клиентская область)
  glViewport(0, 0, w, h);
}

/// перерисовка сцены
void repaint()
{
  // очистить окно, закрасив его цветом очистки
  glClear(GL_COLOR_BUFFER_BIT);
  scene.paint();

  glFinish(); // дождаться выполнения всех OpenGL команд
  glutSwapBuffers(); // обменять буферы кадра, чтобы вывести в окно новый кадр (двойная буферизация)
}

/// обновить состояние сцены
void frame_move()
{
  // "спать" 8 миллисекунд, ограничивая таким образом перерисовку 1000/8 = 125 кадрами в секунду
  avo::sleep(0.008); // на деле это может быть далеко не 8мс, а как захочет ОС :)
  const double t1 = avo::seconds(); // новый момент времени
  scene.move((float)(t1 - t0)); // шаг
  t0 = t1;
  glutPostRedisplay(); // попросим GLUT выполнить перерисовку
}

/// обработать нажатие "обычной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_key(unsigned char key, int x, int y)
{
  switch(key)
  {
  case 27: case 'q': case 'Q':
    glutLeaveMainLoop(); // покинуть цикл обработки сообщений
    break;

  case ' ': case '\r': case '\n':
    scene.make_random();
    break;

  case '-':
    scene.resize(scene.size() - 1);
    break;

  case '=': case '+':
    scene.resize(scene.size() + 1);
    break;

  case '1': scene.mode(GL_POINTS);          break;
  case '2': scene.mode(GL_LINES);           break;
  case '3': scene.mode(GL_LINE_STRIP);      break;
  case '4': scene.mode(GL_LINE_LOOP);       break;
  case '5': scene.mode(GL_TRIANGLES);       break;
  case '6': scene.mode(GL_TRIANGLE_STRIP);  break;
  case '7': scene.mode(GL_TRIANGLE_FAN);    break;
  case '8': scene.mode(GL_QUADS);           break;
  case '9': scene.mode(GL_QUAD_STRIP);      break;

  case '\b':
    glShadeModel( (smooth_shading = !smooth_shading)? GL_SMOOTH: GL_FLAT);
    break;

  default:
    return; // остальные коды игнорируются
  }
}

/// обработать нажатие "специальной" клавиши (x, y -- координаты мыши в окне на момент нажатия)
void on_special(int key, int x, int y)
{
  switch (key)
  {
    // выбор режима растеризации полигонов
  case GLUT_KEY_F10: glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);  break; // точки
  case GLUT_KEY_F11: glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);   break; // линии
  case GLUT_KEY_F12: glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);   break; // нормальный (закрашивать)
  default:
    return; // остальные коды игнорируются
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv); // обработка параметров командной строки
  std::srand((unsigned)(avo::seconds() * 1e+6) % 15485863);
  smooth_shading = true;

  // создать окно с привязанным контекстом OpenGL
  // режим: 32-битный цвет RGBA, двойная буферизация DOUBLE, попытаться использовать полноэкранное сглаживание
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  glutInitWindowPosition(0, 0); // координаты верхнего левого угла окна на экране
  glutInitWindowSize(800, 600); // полные ширина и высота окна
  glutCreateWindow("GLPrimitivesVA");  // создать окно, в функцию передаётся заголовок создаваемого окна

  // параметры OpenGL для данного примера
  glEnable(GL_COLOR_MATERIAL); // выбирать цвет материала командой glColor (через Color::select)
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE); // этот цвет -- с обеих сторон полигона, ambient и diffuse компоненты
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f); // установить цвет очистки (фон) -- чёрный
  glPointSize(5.f); // установить диаметр точки -- 5 пикселей
  glEnable(GL_POINT_SMOOTH); // включить "сглаживание точек" (отдельно от полноэкранного сглаживания)
    // -- без этой команды на некоторых устройствах/версиях драйверов точки будут отображаться не кружками, а квадратами
  glEnable(GL_LINE_SMOOTH);  // включить "сглаживание линий" (отдельно от полноэкранного сглаживания)
 
  // включить рисование примитивов, заданных массивами
  glEnableClientState(GL_COLOR_ARRAY);  // включить массив цветов вершин
  glEnableClientState(GL_VERTEX_ARRAY); // включить массив координат вершин

  // привязка функций-обработчиков событий
  glutIdleFunc(frame_move);      // вызывать, когда нет событий
  glutReshapeFunc(reshape);     // изменение размеров окна
  glutDisplayFunc(repaint);     // вызывать для отрисовки кадра
  glutKeyboardFunc(on_key);     // вызывать для обработки нажатий "обычных" клавиш
  glutSpecialFunc(on_special);  // вызывать для обработки нажатий "специальных" клавиш

  // старт
  scene.resize(24);
  scene.make_random();
  t0 = avo::seconds();
  glutMainLoop(); // вход в цикл обработки событий
  return 0;
}
