﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file  Contour.hpp
 * @brief ориентированный двумерный контур (многоугольник)
 */
#pragma once

#include <vector>
#include <memory>
#include <istream>
#include <ostream>

#include <avo/Vec2.hpp>
#include <avo/Bbox2.hpp>


/// тип скаляра
using Scalar = double;
/// тип вектора, используемый в GLUPoly
using Vec_2 = avo::Vec_2<Scalar>;
/// прямоугольник со сторонами, параллельными осям координат
using Bbox_2 = avo::Bbox_2<Scalar>;

/// ориентированный двумерный контур, основан на std::vector
/**
 * Расширение используется для того, чтобы "примешать" к вектору дополнительные операции,
 * но так как std::vector не предназначен служить базой наследования, используется private-расширение.
 */
class Contour : private std::vector<Vec_2>
{ 
  using Base = std::vector<Vec_2>;

  Base& to_base() { return *this; }
  const Base& to_base() const { return *this; }

public:

  // то, что ожидает увидеть пользователь std::vector
  using Base::iterator;
  using Base::const_iterator;
  using Base::reverse_iterator;
  using Base::const_reverse_iterator;
  
  using Base::allocator_type;
  using Base::value_type;
  using Base::reference;
  using Base::const_reference;
  using Base::pointer;
  using Base::const_pointer;
  using Base::difference_type;
  using Base::size_type;

  using Base::begin;
  using Base::end;
  using Base::cbegin;
  using Base::cend;
  using Base::rbegin;
  using Base::rend;
  using Base::crbegin;
  using Base::crend;

  using Base::size;
  using Base::max_size;
  using Base::resize;
  using Base::capacity;
  using Base::empty;
  using Base::reserve;
  using Base::data;
  using Base::shrink_to_fit;
    
  using Base::operator[];
  using Base::at;
  using Base::front;
  using Base::back;

  using Base::assign;
  using Base::push_back;
  using Base::emplace_back;
  using Base::pop_back;
  using Base::insert;
  using Base::emplace;
  using Base::erase;
  using Base::clear;
  using Base::get_allocator;

  // конструкторы и присваивание

  Contour() {}
  explicit Contour(size_type n)
    : Base(n) {}

  Contour(size_type n, const value_type &val)
    : Base(n, val) {}

  Contour(const Contour &other)
    : Base(other) {}

  Contour(Contour &&other)
    : Base(std::move(other.to_base())) {}

  template <class T>
  Contour(std::initializer_list<T> inl)
    : Base(inl) {}

  template <class InIt>
  Contour(InIt from, InIt to)
    : Base(from, to) {}

  Contour& operator=(const Contour &other)
  {
    to_base() = other;
    return *this;
  }

  Contour& operator=(Contour &&other)
  {
    to_base() = std::move(other.to_base());
    return *this;
  }

  template <class T>
  Contour& operator=(std::initializer_list<T> inl)
  {
    assign(inl.begin(), inl.end());
    return *this;
  }

  // прочие операции

  /// обмен
  void swap(Contour &other)
  {
    Base::swap(other.to_base());
  }

  /// вычислить ориентированную площадь
  Scalar area() const;

  /// вычислить периметр
  Scalar perimeter() const;

  /// сгенерировать bbox по вершинам контура (O(N) operation)
  Bbox_2 box() const
  {
    return Bbox_2(begin(), end());
  }

  // сравнения

  bool operator==(const Contour &other) const
  {
    return to_base() == other.to_base();
  }

  bool operator!=(const Contour &other) const
  {
    return !(*this == other);
  }

  /// параллельный сдвиг
  Contour& operator+=(const Vec_2 &offset)
  {
    for (auto &x : *this)
      x += offset;
    return *this;
  }

  /// параллельный сдвиг
  Contour& operator-=(const Vec_2 &offset)
  {
    return *this += (-offset);
  }

  /// изотропное масштабирования
  Contour& operator*=(Scalar scale)
  {
    for (auto &x : *this)
      x *= scale;
    return *this;
  }

  /// анизотропное масштабирование
  Contour& operator*=(const Vec_2 &scale_xy)
  {
    for (auto &x : *this)
      x *= scale_xy;
    return *this;
  }
};


/// swap function
inline void swap(Contour &a, Contour &b)
{
  a.swap(b);
}


inline std::ostream& operator<<(std::ostream &os, const Vec_2 &p)
{
  return os << p.x << ' ' << p.y;
}

inline std::istream& operator>>(std::istream &is, Vec_2 &p)
{
  return is >> p.x >> p.y;
}


std::ostream& operator<<(std::ostream &os, const Contour&);
std::istream& operator>>(std::istream &is, Contour&);
