/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file   Bbox2.hpp
 * @brief  2D axes-aligned bounding box (rectangle)
 * @author Dmitri Kuvshinov
 *
 * The problem with the current Bbox2 is that it doesn't have
 * any means to distinguish "empty" box from the box containing just one point.
 */
#pragma once
#include <limits>  // epsilon used in forceAspect
#include <avo/Vec2.hpp>


namespace avo
{

/// 2D axes-aligned bounding box
template <class Float>
struct Bbox_2
{
  using Vec  = Vec_2<Float>;
  using This = Bbox_2<Float>;

  Vec pmin ///< left-lower point
    , pmax ///< right-upper point
    ;

  /// origin
  Bbox_2() {}
  
  /// a point
  explicit Bbox_2(const Vec &p)
    : pmin(p), pmax(p) {}

  /// minimal bbox containing both a and b
  Bbox_2(const Vec &a, const Vec &b)
    : pmin(a), pmax(b)
  {
    pmax.maximize(a);
    pmin.minimize(b);
  }

  /// minimal bbox containing all points from the memory range
  /**
   * Should come out an optimized version against InputIterator-based template analogue below.
   */
  Bbox_2(const Vec *begin, const Vec *end)
    : pmin(Vec()), pmax(Vec())
  {
    bbox2(begin, end - begin, pmin, pmax);
  }

  /// minimal bbox containing all points from the range
  template <class InIt>
  Bbox_2(InIt begin, InIt end)
  {
    if (begin != end)
    {
      pmax = pmin = *begin;
      while (++begin != end)
        grow(*begin);
    }
  }


  /// checks if the box has no area ("is empty")
  bool empty() const
  {
    return pmin == pmax;
  }

  /// checks if pmax is not dominated by pmin
  bool valid() const
  {
    return !pmin.dominates(pmax);
  }

  /// checks if the given point falls inside the box
  bool contains(const Vec &p) const
  {
    return p.dominates(pmin) && pmax.dominates(p);
  }

  /// checks if the given box is inside this box
  bool contains(const This &b) const
  {
    return contains(b.pmin) && contains(b.pmax);
  }

  /// width of the box
  Float width() const
  {
    return pmax.x - pmin.x;
  }

  /// height of the box
  Float height() const
  {
    return pmax.y - pmin.y;
  }

  /// calculate width and height simultaneously
  Vec extents() const
  {
    return pmax - pmin;
  }

  /// width divided by height
  /**
   * Undefined result if height() == 0.0.
   */
  Float aspect() const
  {
    return width() / height();
  }

  /// move the box by the given delta vector
  This& move(const Vec &delta)
  {
    pmin += delta;
    pmax += delta;
    return *this;
  }

  /// grow the box to contain the given point
  This& grow(const Vec &p)
  {
    pmin.minimize(p);
    pmax.maximize(p);
    return *this;
  }

  /// grow the box to contain the given box
  This& grow(const This &other)
  {
    pmin.minimize(other.pmin);
    pmax.maximize(other.pmax);
    return *this;
  }

  /// intersect with another box, may become invalid when intersection is empty
  This& intersect(const This &other)
  {
    pmin.maximize(other.pmin);
    pmax.minimize(other.pmax);
    return *this;
  }

  /// calculates area of the box
  Float area() const
  {
    const Vec ext = extents();
    return ext.x * ext.y;
  }


  /// box of the union of this and other boxes
  This& operator+=(const This &other)
  {
    return grow(other);
  }

  /// intersection of two boxes
  This& operator&=(const This &other)
  {
    return intersect(other);
  }


  /// make the box containing this but providing the aspect (width / height) needed
  void force_aspect(Float aspect)
  {
    // ignore NaN, zero, infinity and negative
    if (aspect <= Float(0) || !(aspect == aspect) || Float(0.5) * aspect == aspect)
      return;

    using std::sqrt;
    // w0, h0 -- old sizes, area == w0 * h0,
    // needed aspect == aspect, so
    // w1 == aspect * h1 && w1 * h1 == w0 * h0
    // ==> h1 == w0 * h0 / w1 ==> w1 * w1 == aspect * w0 * h0
    const Float w0 = width(), h0 = height(), area = w0 * h0,
                w1 = sqrt(aspect * area);

    if (Float(0.5) * w1 == w1)
      return;

    const Float h1 = area / w1;
    if (Float(0.5) * h1 == h1)
      return;

    const auto gv = Float(0.5) * Vec(w1 - w0, h1 - h0);
    pmax += gv;
    pmin -= gv;
  }
};


template <class Float>
inline bool operator==(const Bbox_2<Float> &a, const Bbox_2<Float> &b)
{
  return a.pmin == b.pmin && a.pmax == b.pmax;
}

template <class Float>
inline bool operator!=(const Bbox_2<Float> &a, const Bbox_2<Float> &b)
{
  return !(a == b);
}


/// build box of the union of two boxes
template <class Float>
inline Bbox_2<Float> operator+(const Bbox_2<Float> &a, const Bbox_2<Float> &b)
{
  return (Bbox_2<Float>(a)) += b;
}

/// build box of the intersection of two boxes
template <class Float>
inline Bbox_2<Float> operator&(const Bbox_2<Float> &a, const Bbox_2<Float> &b)
{
  return (Bbox_2<Float>(a)) &= b;
}

}
