/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Camera2.hpp
 * @brief 2D "camera" implementation (controllable bbox2)
 * @author Dmitri Kuvshinov
 */
#pragma once
#include <GL/freeglut.h>
#include <avo/Bbox2.hpp>


namespace avo
{

/// 2D camera representation
class Camera_2
{
public:
  using Box = Bbox_2<double>;
  using Vec = Box::Vec;

private:
  Box repr;

public:

  /// maps to (-1.0, -1.0) - (1.0, 1.0) by default
  Camera_2()
    : repr(Vec(-1.0, -1.0), Vec(1.0, 1.0)) {}


  /// get the box projected to the viewport
  const Box& box() const
  {
    return repr;
  }

  /// move in plane
  void move(const Vec &delta)
  {
    repr.move(delta);
  }

  /// promote viewport aspect (width to height ratio)
  void force_aspect(double aspect)
  {
    repr.force_aspect(aspect);
  }

  /// zoom camera in, factor == 1 will shrink the box into a point, factor == 0.5 will halve it in both dimensions
  void zoom_in(double factor)
  {
    const Vec shr = (0.5 * factor) * repr.extents();
    repr.pmin += shr;
    repr.pmax -= shr;
  }

  /// zoom camera out, factor == 1 will double the box in both dimensions
  void zoom_out(double factor)
  {
    const Vec grw = (0.5 * factor) * repr.extents();
    repr.pmin -= grw;
    repr.pmax += grw;
  }


  /// apply camera to OpenGL state, leaves MODELVIEW matrix selected
  void apply() const
  {
    // setup the corresponding ortho
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(repr.pmin.x, repr.pmax.x, repr.pmin.y, repr.pmax.y);
    glMatrixMode(GL_MODELVIEW);
  }

  /// reset camera according to scene contents, keeping ascpect unchanged
  void reset(const Box &target)
  {
    if (!target.empty())
    {
      const auto aspect = repr.aspect();
      repr = target;
      repr.force_aspect(aspect);
    }
  }
};

}
