/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file SphericalCamera.hpp
 * @author Dmitri Kuvshinov
 * @brief spherical coordinates to simulate camera rotating around a point at which it's looking
 * @todo change implementation to matrix?
 */
#pragma once
#include <GL/freeglut.h> // for OpenGL functions
#include <avo/Vec3.hpp>
#include <cmath>


namespace avo
{

const double PI       = 3.14159265358979323846;
const double PI_div_2 = 1.57079632679489661923;
const double PI_mul_2 = 2.0 * PI;


/// "camera" in spherical coordinates
class Spherical_camera
{
  double phi, theta, r; // spherical coordinates
  double dphi, dtheta;  // sphere angle steps
  double rf;            // r increase factor

  // precomputed sines and cosines
  double cos_phi, sin_phi, cos_theta, sin_theta;

  // sphere center
  Vec_3d orig;

public:
  /// height of eye over Oxy plane
  double z_height() const
  {
    return r * cos_theta;
  }

  /// get actual camera position
  Vec_3d position() const
  {
    return Vec_3d{ r * sin_theta * cos_phi, r * sin_theta * sin_phi, z_height() }
         + orig;
  }

  /// initialize step values
  explicit Spherical_camera
    (
      unsigned  longitude_steps = 90
    , unsigned  latitude_steps  = 50
    , double    distance_factor = 0.9
    , Vec_3d    orig = Vec_3d{}
    )
      : dphi   (PI_mul_2 / longitude_steps)
      , dtheta (PI / latitude_steps)
      , rf     (distance_factor)
      , phi       (0.0) , theta     (PI_div_2) , r (500.0)
      , cos_phi   (1.0) , sin_phi   (0.0)
      , cos_theta (0.0) , sin_theta (1.0)
      , orig(orig)
  {}

  /// constructor from coordinates
  Spherical_camera
    (
      double    phi
    , double    theta
    , double    r
    , unsigned  longitude_steps = 90
    , unsigned  latitude_steps  = 50
    , double    distance_factor  = 0.9
    , Vec_3d    orig = Vec_3d{}
    )
      : dphi   (PI_mul_2 / longitude_steps)
      , dtheta (PI / latitude_steps)
      , rf     (distance_factor)
      , phi       (phi)             , theta     (theta)           , r (r)
      , cos_phi   (std::cos(phi))   , sin_phi   (std::sin(phi))
      , cos_theta (std::cos(theta)) , sin_theta (std::sin(theta))
      , orig(orig)
  {}


  /// rotate right
  void right()
  {
    phi += dphi;
    if (PI_mul_2 < phi) phi -= PI_mul_2;

    cos_phi = std::cos(phi);
    sin_phi = std::sin(phi);
  }

  /// rotate left
  void left()
  {
    phi -= dphi;
    if (phi < 0.0) phi += PI_mul_2;

    cos_phi = std::cos(phi);
    sin_phi = std::sin(phi);
  }

  /// rotate up
  void up()
  {
    theta -= dtheta;
    if (theta < 0.0) theta += PI_mul_2;
    
    cos_theta = std::cos(theta);
    sin_theta = std::sin(theta);
  }

  /// rotate down
  void down()
  {
    theta += dtheta;
    if (PI_mul_2 < theta) theta -= PI_mul_2;

    cos_theta = std::cos(theta);
    sin_theta = std::sin(theta);
  }

  /// move forward
  void forward()
  {
    r *= rf;
  }

  /// move backward
  void backward()
  {
    r /= rf;
  }

  /// get current origin value
  const Vec_3d& origin() const
  {
    return orig;
  }

  /// move origin by the delta vector
  void move_origin(const Vec_3d &delta)
  {
    orig += delta;
  }

  /// get camera viewing direction (unit vector)
  Vec_3d direction() const
  {
    return Vec_3d{ -sin_theta * cos_phi, -sin_theta * sin_phi, -cos_theta };
  }


  // origin movement according to the camera direction of view
  /// move origin forward dist units, use -dist to move backward
  void origin_forward(double dist)
  {
    move_origin(dist * direction());
  }

  /// move origin right dist units, use -dist to move left
  void origin_right(double dist)
  {
    // pseudoup vector is (0.0, 0.0, (PI < theta? -1.0: 1.0))
    /*
      ex ey ez
      0  0  s
      dx dy dz
      => -s dy ex + s dx ey + 0 ez
    */
    if (PI > theta) dist = -dist;
    move_origin((dist * sin_theta) * Vec_3d{ sin_phi, -cos_phi }.normalize());
  }

  /// move origin "up" dist units, use -dist to move down
  void origin_up(double dist)
  {
    // pseudoup vector is (0.0, 0.0, (PI < theta? -1.0: 1.0))
    // we should find a perpendicular to direction closest to this pseudoup in the same plane (dir, pseudoup)
    Vec_3d up{ 0., 0., (PI < theta? -1.0: 1.0) };
    const auto dir = direction(); // has unit length
    up -= dot(up, dir) * dir;
    up.normalize() *= dist;
    move_origin(up);
  }

  /// apply the camera matrix to the projection matrix, which is placed second on the matrix stack
  void apply() const
  {
    glMatrixMode(GL_PROJECTION);

    glPopMatrix();  // duplicate the previous matrix on the stack
    glPushMatrix();
    const auto pos = position();
    gluLookAt
      (
        pos.x          // camera position x
      , pos.y          // camera position y
      , pos.z          // camera position z
      , orig.x, orig.y, orig.z  // camera looks at origin
      , 0.0, 0.0, (PI < theta? -1.0: 1.0)  // camera's up vector
      );

    glMatrixMode(GL_MODELVIEW);
  }
};

}
