﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2014, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file  TextDrawer.hpp
 * @brief класс, инкапсулирующий средства рисования текста
 */
#pragma once
#include <GL/freeglut.h>
#include "Bbox2.hpp"

#include <cassert>
#include <string>
#include <algorithm>


namespace avo
{

/// аналог glRasterPos2d, смещающий текущую позицию на заданные сдвиги по oX и по oY в пикселях
inline void moveWindowPos(double dx, double dy)
{
  glBitmap(0, 0, 0, 0, (GLfloat)dx, (GLfloat)dy, nullptr);
}


/// положение текста относительно текущей растровой позиции (флаги -- можно объединять с помощью |)
namespace Alignment
{
  enum Type
  {
    toUp      = 0x0,
    toRight   = 0x0,
    toDown    = 0x1,
    toLeft    = 0x2,
    centerX   = 0x4,
    centerY   = 0x8
  };
}

/// Класс, инкапсулирующий средства вывода текста FreeGlut.
/// Предполагает ортогональную проекцию с обычной системой координат (oY -- вверх, oX -- вправо)
class Bitmap_text_drawer_2
{
public:
  using This  = Bitmap_text_drawer_2;
  using Bbox_2 = avo::Bbox_2<double>;
  using Vec_2  = Bbox_2::Vec;


  Bitmap_text_drawer_2()
    : cur_font(GLUT_BITMAP_HELVETICA_10),
      cur_wnd(Vec_2(0, 0), Vec_2(1, 1)),
      cur_ortho(Vec_2(0, 0), Vec_2(1, 1)),
      cur_wcoef(0.),
      cur_hcoef(0.)
  {}

  /// выбрать другой шрифт
  This& font(void *another_font)
  {
    cur_font = another_font;
    return *this;
  }

  /// выбрать новую позицию (в координатах мира)
  This& view(double x, double y)
  {
    glRasterPos2d(x, y);
    return *this;
  }

  /// привести пиксельные координаты к координатам мира
  void pixels_to_view(int x, int y, double &vx, double &vy) const
  {
    Vec_2 r(x, y);
    r -= cur_wnd.pmin;
    r.x *= cur_wcoef;
    r.y *= cur_hcoef;
    r += cur_ortho.pmin;
    vx = r.x;
    vy = r.y;
  }

  /// выбрать новую позицию (в пиксельных координатах окна)
  This& at(int x, int y)
  {
    double vx, vy;
    pixels_to_view(x, y, vx, vy);
    return view(vx, vy);
  }

  /// вывести текст (смещает текущую позицию вывода)
  This& put(const std::string &text, int alignment = Alignment::toUp | Alignment::toRight)
  {
    apply_alignment(text, alignment);
    glutBitmapString(cur_font, (const unsigned char*)text.c_str());
    return *this;
  }

  /// получить размеры текста от FreeGlut (в пикслелях)
  /// текст может быть многострочным, но междустрочный и межбуквенный интервалы фиксированы
  void pixel_text_extents(const std::string &text, double &width, double &height) const
  {
    height = (1 + std::count(text.begin(), text.end(), '\n')) * (double)glutBitmapHeight(cur_font);
    width = (double)glutBitmapLength(cur_font, (const unsigned char*)text.c_str());
  }

  /// размеры текста в координатных единицах текущей ортогональной проекции
  void ortho_text_extents(const std::string &text, double &width, double &height) const
  {
    assert(cur_wcoef != 0. && cur_hcoef != 0.); // забыли установить размеры окна set_viewport и set_coordinates
    pixel_text_extents(text, width, height);
    width  *= cur_wcoef;
    height *= cur_hcoef;
  }

  // необходимо уведомлять класс о происходящих изменениях проекции или окна
  /// изменение окна просмотра
  void set_viewport(int x0, int y0, int width, int height)
  {
    assert(width != 0 && height != 0);
    cur_wnd.pmin = Vec_2(x0, y0);
    cur_wnd.pmax = cur_wnd.pmin + Vec_2(width, height);
    recompute_whcoef();
  }

  /// изменение окна просмотра (x0 == 0 && y0 == 0)
  void set_viewport(int width, int height)
  {
    set_viewport(0, 0, width, height);
  }

  /// изменение проекции
  void set_coordinates(const Bbox_2 &ortho)
  {
    cur_ortho = ortho;
    recompute_whcoef();
  }

private:
  void *cur_font;
  Bbox_2 cur_ortho, cur_wnd;
  double cur_wcoef, cur_hcoef;

  /// пересчитать коэффициенты масштабирования размеров текста под текущую систему координат
  void recompute_whcoef()
  {
    assert(cur_wnd.valid() && cur_ortho.valid());
    cur_wcoef = cur_ortho.width() / cur_wnd.width();
    cur_hcoef = cur_ortho.height() / cur_wnd.height();
  }

  /// поправить текущую растровую позицию под заданное положение текста
  void apply_alignment(const std::string &text, int al) const
  {
    using namespace Alignment;
    if (al == 0)
      return;

    double w, h;
    pixel_text_extents(text, w, h);
    
    if (al & toLeft)
      w = -w;
    else if (al & centerX)
      w *= -0.5;
    else
      w = 0.;

    if (al & toDown)
      h = -h;
    else if (al & centerY)
      h *= -0.5;
    else
      h = 0.;

    moveWindowPos(w, h);
  }
};

}
