﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * Пример: SortVis
 * Автор: Кувшинов Д.Р. https://evetro.wordpress.com/
 * Используемые библиотеки: FreeGlut, OpenGL.
 *
 * Визуализатор различных алгоритмов сортировки.
 * Также см. https://evetro.wordpress.com/2011/04/28/av-1/
 * Нажатие:
 * +/-   -- ускорение/замедление демонстрации;
 * ENTER -- ввод параметров из консоли:
 *   скорость (количество событий в секунду),
 *   количество,
 *   набор алгоритмов, заканчивающийся точкой с запятой (просто ; для повтора);
 * 1, 2, 3 ... -- выбор типа последовательности;
 * SPACE -- перезапуск процесса сортировки;
 * P     -- пауза;
 * T     -- включить/выключить отображение названий алгоритмов;
 * ESC/Q -- выход из программы.
 */
#include <GL/freeglut.h>
#include <cassert>
#include <cstdlib>     // rand/srand
#include <string>
#include <iostream>
#include <iomanip>
#include <vector>      // динамический массив
#include <algorithm>   // std::sort + некоторые вспомогательные элементы
#include <functional>  // std::mem_fn, std::function
#include <numeric>     // std::accumulate
#include <iterator>    // std::iterator_traits
#include <map>         // словарь алгоритмов сортировки и очередь действий
#include <queue>
#include <utility>     // std::move

#include <avo/Color.hpp>
#include <avo/Timer.hpp>
#include "Sorts.hpp"     // реализации алгоритмов сортировки
#include "Sequences.hpp" // реализации генераторов последовательностей

using avo::Color;
using avo::RED;
using avo::ORANGE;
using avo::GREEN;
using avo::GREY;
using avo::WHITE;
using avo::YELLOW;

using avo::clamp;


//
// вспомогательные определения
//

/// Отложенное действие, всё, что с ним можно сделать -- "выполнить", применив ().
using Action  = std::function<void()>;
/// Контейнер отложенных действий.
using Actions = std::vector<Action>;

/// Очередь событий ("действий"), упорядоченных по временной метке.
/** Action_queue ничего не знает о природе хранимых в ней действий */
class Action_queue
{
public:
  /// Добавить новый объект действия с заданной временной меткой.
  void push(double time, Action action)
  {
    aq.emplace(time, std::move(action));
  }

  /// Состояние очереди после извлечения очередной порции отложенных действий.
  enum Status { Empty, Added_nothing, Added_some };

  /// Извлечь все отложенные действия с временными метками не больше переданной (перемещаются в actions).
  /** возвращает истину, если какие-либо действия были выполнены */
  Status next(double time, Actions &actions)
  {
    if (aq.empty())
      return Empty;
    if (time < aq.top().stamp)
      return Added_nothing;

    while (true)
    {
      actions.emplace_back(std::move(aq.top().action));
      aq.pop();
      if (aq.empty() || time < aq.top().stamp)
        return Added_some;
    }
  }

private:
  struct Pair
  {
    double stamp;
    Action action;

    Pair() : stamp(0.0) {}
    Pair(double t, Action &&a)
      : stamp(t), action(std::move(a)) {}

    bool operator<(const Pair &other) const
    {
      return other.stamp < stamp;
    }
  };

  std::priority_queue<Pair> aq;
};


/// набор строк
using Strings = std::vector<std::string>;


//
// сцена
//

/// прямоугольник со сторонами, параллельными осям координат, который можно нарисовать
struct Rect // визуализирует элемент сортируемой последовательности
{
  float x0, y0; // координаты нижнего левого угла
  float x1, y1; // координаты верхнего правого угла
  Color color;  // цвет заливки прямоугольника

  /// базовый конструктор
  explicit Rect(float x0 = 0.f, float y0 = 0.f,
                float x1 = 0.f, float y1 = 0.f, 
                const Color &col = GREY)
      : x0(x0), y0(y0), x1(x1), y1(y1), color(col)
  {}

  /// OpenGL1-код, рисующий данный прямоугольник
  void paint() const
  {
    color.select();
    glRectf( x0, y0, x1, y1 );
  }
};


/// массив прямоугольников, который можно нарисовать
class Rect_array
{
  using Rects = std::vector<Rect>;
  Rects rects;  // набор прямоугольников
  float dx, dy; // сдвиг массива в плоскости

public:
  Rect_array() {}

  /// сгенерировать набор прямоугольников, длины которых равны числам из диапазона [begin, end)
  template <class FwdIt>
  Rect_array(FwdIt begin, FwdIt end, float dx = 0.0f, float dy = 0.0f)
    : dx(dx), dy(dy)
  {
    // заполнить rects
    while (begin != end)
      rects.emplace_back(0.f, 0.f, *begin++);

    // сместить прямоугольники
    const unsigned sz = rects.size();
    for (unsigned i = 0; i < sz; ++i)
    {
      Rect &r = rects[i];
      r.y0 = r.y1 = float(i);
      r.y1 += 0.75f;
    }
  }

  /// указать цвет для всех прямоугольников
  void clear_color(const Color &col = GREY)
  {
    for (Rect &r: rects)
      r.color = col;
  }

  /// нарисовать все прямоугольники
  void paint() const
  {
    glLoadIdentity();
    glTranslatef( dx, dy, 0.0f );
    for (const Rect &r: rects)
      r.paint();
  }


  // реакция на события

  /// событие "присваивание": [index] = value
  void assign(unsigned index, float value)
  {
    Rect &r = rects.at(index);
    r.color = RED;
    r.x1 = value;
  }

  /// событие "сравнение": [i1] < [i2]?
  void compare(unsigned i1, unsigned i2)
  {
    rects.at(i1).color = GREEN;
    rects.at(i2).color = GREEN;
  }

  /// событие "обмен": [i1] <-> [i2]
  void swap(unsigned i1, unsigned i2)
  {
    rects.at(i1).color = ORANGE;
    rects.at(i2).color = ORANGE;
    std::swap( rects[i1].x1, rects[i2].x1 );
  }
};


/// Time_mapper формирует временные метки событий и служит связующим звеном между набором зондов и очередью событий
class Time_mapper
{
private:
  Rect_array   *ra; // связанный массив прямоугольников
  Action_queue *eq; // связанная очередь событий
  double t, dt;     // начальный момент времени и шаг

  // статистика
  std::size_t stat_assign, stat_comp, stat_swap;

public:
  // для создания объекта требуется передать объект массива прямоугольников и объект очереди событий
  Time_mapper(Rect_array &R, Action_queue &Q, double time = 0., double step = 1.)
    : ra(&R), eq(&Q), t(time), dt(step), stat_assign(0), stat_comp(0), stat_swap(0)
  {}

  /// добавить действие и вычислить следующую временную метку
  void push_next(Action action)
  {
    eq->push(t, std::move(action));
    t += dt;
  }

  // Time_mapper "знает", какие виды событий могут быть,
  // это не очень приятно, но альтернатива -- хранить указать ra прямо в зондах -- мне не очень по душе.

  void push_assignment(unsigned i, float v)
  {
    ++stat_assign;
    push_next(std::bind(&Rect_array::assign, ra, i, v));
  }

  void push_comparison(unsigned i1, unsigned i2)
  {
    ++stat_comp;
    push_next(std::bind(&Rect_array::compare, ra, i1, i2));
  }

  void push_swap(unsigned i1, unsigned i2)
  {
    ++stat_swap;
    push_next(std::bind(&Rect_array::swap, ra, i1, i2));
  }

  // извлечение статистики
  std::size_t assignments() const { return stat_assign; }
  std::size_t comparisons() const { return stat_comp; }
  std::size_t swaps() const { return stat_swap; }
};


/// зонд генерирует события в процессе работы алгоритмов сортировки
/**
 * Таким образом, используя "зонды" вместо обычных объектов, можно записывать
 * проделываемые на них операции, чтобы потом визуализировать их.
 */
class Probe
{
  // объект Time_mapper служит связующим звеном между набором зондов, очередью событий
  Time_mapper *tm; // и массивом прямоугольников
  unsigned index; // позиция зонда в массиве
  float value;    // числовое значение зонда (длина прямоугольника)

public:
  /// базовый конструктор
  explicit Probe(unsigned i = 0, float v = 0.f, Time_mapper *tm = nullptr)
    : tm(tm), index(i), value(v)
  {}

  /// сгенерировать событие "присваивание"
  Probe& operator=(const Probe &right)
  {
    if (tm) tm->push_assignment(index, right.value);
    value = right.value;
    return *this;
  }

  /// сгенерировать событие "сравнение"
  bool operator<(const Probe &right) const
  {
    if (tm) tm->push_comparison(index, right.index);
    return value < right.value;
  }

  // Если требуется сравнивать на равенство,
  // добавить сюда operator==, аналогичный operator<, определённому выше
  // -- тоже событие "сравнение".

  /// сгенерировать событие "обмен"
  void swap(Probe &other)
  {
    if (tm) tm->push_swap(index, other.index);
    std::swap(value, other.value);
  }
};

/// выбирается вместо базовой std::swap благодаря ADL
inline void swap(Probe &a, Probe &b) 
{
  a.swap(b);
}


/// сцена представляет собой фасад всего "безобразия", описанного выше :)
class Scene
{
private:
  using Probes = std::vector<Probe>;
  using PIt    = Probes::iterator;
  // параллельные массивы:
  std::vector<Action_queue> queues;     // очереди событий
  std::vector<Rect_array>   arrays;     // сортируемые массивы прямоугольников, отвечающие за визуализацию
  Strings                   algorithms; // названия алгоритмов сортировки
  Strings                   titles;     // надписи (названия алгоритмов, возможно, номера их завершения)
  std::vector<size_t>       stopped;    // порядок завершения сортировок (0 -- "не завершена") в процессе визуализации
  size_t already_stopped; // сколько сортировок уже завершились

  // выбор генератора последовательности
  using Samples = std::vector<float>;
  using SIt     = Samples::iterator;
  using Generator_function = std::function<void(SIt, SIt)>;
  Generator_function generator;

public:
  Scene()
    : height(1.f),
      generator(generate_random<SIt>),
      is_paused(false),
      show_titles(true),
      last_time(0.0),
      time_offset(0.0),
      already_stopped(0)
  {
    fill_sorts(); // сформировать базовый набор алгоритмов сортировки,
    // чтобы можно было выбирать алгоритм по названию
  }

  /// сбрасывает внутренние структуры, заполняет массивы случайной последовательностью, сортирует
  /**
   * Принимает список названий алгоритмов (строки), заданный диапазоном [algorithm_begin, algorithm_end).
   * Названия см. в fill_sorts.
   */
  template <class It>
  void resort(unsigned amount, It algorithm_begin, It algorithm_end)
  {
    using namespace std;
    height = float(amount); // высота одного столбца (используется при рисовании)

    Strings old_algorithms;
    algorithms.swap(old_algorithms); 

    // теперь algorithms пуст, а старый набор алгоритмов сохранён в old_algorithms,
    // оставить только те алгоритмы, которые есть в словаре:
    copy_if (algorithm_begin, algorithm_end, back_inserter(algorithms),
      [this](const string &name)
      {
        return sorts.count(name) != 0;
      });

    // если нет годных алгоритмов -- вернуть старый набор:
    if (algorithms.empty())
      algorithms.swap(old_algorithms);

    // если набор алгоритмов оказался пуст -- использовать набор алгоритмов по умолчанию:
    if (algorithms.empty())
      default_algorithms();

    arrays.clear();
    queues.clear();
    // инициализировать массивы прямоугольников и очереди событий
    const unsigned  N = algorithms.size();
    arrays.reserve (N);
    queues.resize  (N);
    
    // вспомогательные элементы
    stopped.assign (N, 0);
    already_stopped = 0;
    titles = algorithms;

    Samples samples (amount); // набор псевдослучайных "семплов"

    cout << "generating samples ... " << flush;
    generator(samples.begin(), samples.end());
    cout << "done\n" << endl;

    // инициализировать массивы и очереди
    for (unsigned i = 0; i < N; ++i)
    {
      // создать новый массив
      arrays.emplace_back(samples.begin(), samples.end(), float(i));

      // выполнить сортировку
      Time_mapper tm(arrays[i], queues[i]);  // связующее звено
      Probes probes; // сортируемые "зонды"
      probes.reserve(amount);
      for (unsigned j = 0; j < amount; ++j)
        probes.emplace_back(j, samples[j], &tm);

      // собственно сортировка
      cout << left << setw(10) << algorithms[i] << " ... " << flush;
      sorts[algorithms[i]] (probes.begin(), probes.end()); // вызов функции из массива
      cout << "done: " << right;

      // вывести статистику
      cout << setw(10) << tm.comparisons() << "C, "
           << setw(10) << tm.assignments() << "A, "
           << setw(10) << tm.swaps() << "S"
           << setw(10) << 
      // вес операций: сравнение = 1, чтение = 3, запись = 6
      // comparison => 2 чтения + сравнение = 7,
      // assignment => 0.5("вероятность")*чтение + запись = 7.5,
      // swap => 0.5("вероятность")*два чтения + две записи = 9.
                std::ceil( 7.0 * tm.comparisons() + 7.5 * tm.assignments() + 9.0 * tm.swaps() )
           << "ops";
      cout << endl;
    }

    cout.put('\n');
  }

  /// распечатать набор встроенных алгоритмов сортировки
  void print_supported_sorts() const
  {
    for (auto &s : sorts)
      std::cout << s.first << '\n';
    std::cout << std::endl;
  }

  /// изобразить все очереди
  void paint() const
  {
    // установить матрицу проекции, область рисования -- прямоугольник размера N * 1
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(-0.05, algorithms.size() + 0.05, -0.125, height + 0.875);
    glMatrixMode(GL_MODELVIEW);

    // нарисовать каждый набор
    for (auto &a : arrays)
      a.paint();

    // надписи
    if (show_titles)
    {
      for (std::size_t i = 0, iend = titles.size(); i != iend; ++i)
      {
        // выбрать цвет текста
        (stopped[i] == 0? WHITE: YELLOW).select();
        // установить позицию левого нижнего угла отображаемой строки
        glLoadIdentity();
        glRasterPos2f(GLfloat(i), height * 0.09f * (3 + i % 8));
        // изобразить строку текста, используя встроенный растровый шрифт
        glutBitmapString(GLUT_BITMAP_HELVETICA_18, 
          reinterpret_cast<const unsigned char*>(titles[i].c_str()));
      }
    }
  }


  /// Возможный результат next: "все очереди пусты" (простой), "нет изменений" (не рисовать), "есть изменения" (рисовать).
  enum Frame_move { Sleeping, Unchanged, Changed };

  /// Применить следующее событие (достигшее времени time) каждой очереди.
  Frame_move next(double time)
  {
    if (!is_paused)
    {
      time -= time_offset;
      last_time = time;

      // выставляется true, если надо обновить экран (последний раз) после завершения (последней) сортировки
      bool report_empty = false;
      // извлечь отложенные действия, сбросить цвета наборов, для которых произошли события
      // queues и arrays -- "параллельные" массивы
      for (size_t i = 0, iend = queues.size(); i != iend; ++i)
      {
        switch (queues[i].next(time, deferred))
        {
        case Action_queue::Empty:
          if (stopped[i] == 0)
          {
            report_empty = true;
            stopped[i] = ++already_stopped;
            (titles[i] += ": ") += std::to_string(already_stopped);
          }
          break;
        case Action_queue::Added_nothing:
          /* do nothing */
          break;
        case Action_queue::Added_some:
          arrays[i].clear_color();
          break;
        default:
          assert(false);
        }
      }

      // выполнить действия
      for (auto &d : deferred)
        d();

      if (deferred.empty() && !report_empty)
        return already_stopped == queues.size()? Sleeping: Unchanged;

      deferred.clear();
      return Changed;
    }

    // на паузе
    time_offset = time - last_time;
    return Sleeping;
  }

  /// выбрать генератор из предопределённого набора по номеру, сюда добавляются новые генераторы последовательностей
  void change_generator(int n)
  {
    static const Generator_function generators[] =
    {
      generate_random            <SIt>,
      generate_sorted            <SIt>,
      generate_antisorted        <SIt>,
      generate_bitonic           <SIt>,
      generate_antibitonic       <SIt>,
      generate_sorted_blocks     <SIt>,
      generate_antisorted_blocks <SIt>
    };

    if (0 <= n && n < sizeof(generators)/sizeof(Generator_function))
      generator = generators[n];
  }

  // пауза -- снять с паузы
  void toggle_pause()
  {
    is_paused = !is_paused;
  }
  
  // включить/выключить отображение названий алгоритмов
  void toggle_titles()
  {
    show_titles = !show_titles;
  }

private:
  // средства обеспечения паузы поверх текущей реализации
  bool is_paused;
  double last_time, time_offset;
  // ---

  bool show_titles; // показывать названия алгоритмов
  float height; // рабочая высота сцены == количеству сортируемых элементов

  using Sort_function = std::function<void(PIt, PIt)>;
  std::map<std::string, Sort_function> sorts; // позволяет выбрать сортировку по названию во время исполнения

  /// сформировать словарь алгоритмов сортировки (sorts), сюда добавляются новые алгоритмы сортировки
  void fill_sorts()
  {
    #ifdef S
    #error Macro redefinition, check the code.
    #endif
    #define S(name, algo) \
      sorts[name] = algo<PIt>
    S("bubble",     bubble_sort);
    S("cocktail",   cocktail_sort);
    S("comb",       comb_sort);
    S("minmax",     minmax_sort);
    S("insert",     insertion_sort);
    S("quick",      quick_sort);
    S("merge",      merge_sort);
    S("merge2",     merge_sort_2);
    S("merge3",     merge_sort_3);
    S("merge4",     merge_sort_4);
    S("rbitonic",   bitonic_sort);
    S("bitonic",    bitonic_sort_2);
    S("oddeven",    oddeven_sort);
    S("pairwise",   pairwise_sort);
    S("heap",       heap_sort);
    S("random",     random_try_sort);
    S("ci",         ci_sort);
    S("combinsert", ci_sort);
    S("cycle",      cycle_sort);
    S("bst",        tree_sort);
    S("tree",       tree_sort);
    S("bbt",        tree_sort_balanced);
    S("balanced",   tree_sort_balanced);
    S("stdheap",    std_heap_sort);
    S("hybheap",    hybrid_heap_sort);
    S("std",        std::sort);
    S("stable",     std::stable_sort);
    #undef S
  }

  /// набор алгоритмов сортировки "по умолчанию"
  void default_algorithms()
  {
    const char* const ss[] =
    { "comb", "quick", "stable", "rbitonic", "merge", "heap", "insert" };
    algorithms.assign(ss, ss + sizeof(ss)/sizeof(char*));
  }

  // накапливаемые отложенные действия
  Actions deferred;
};


/// "скорость" таймера, хранит множитель
class Rate
{
  double factor;

public:
  explicit Rate(double f = 1.0)
    : factor(f) {}

  void multiply(double f)
  {
    factor *= f;
  }

  double operator()(double dt) const
  {
    return dt * factor;
  }
};


//
// глобальные переменные
//

/// количество сортируемых элементов
unsigned elements_to_be_sorted = 64;

/// используемые алгоритмы сортировки
Strings algorithms;

/// таймер очередей
avo::Timer<Rate> timer;

/// объект сцены
Scene scene;


//
// обработчики событий GLUT
//

/// перерисовка
void repaint()
{
  glClear(GL_COLOR_BUFFER_BIT);
  scene.paint();
  glFinish();
  glutSwapBuffers();
}


/// вычислить состояние сцены на следующий кадр
void frame_move()
{
  switch (scene.next(timer.mark()))
  {
  case Scene::Sleeping:  // режим "простой", ждать 0.02с
    avo::sleep(0.02);
    break;
  case Scene::Unchanged: // нет изменений, не перерисовывать
    break;
  case Scene::Changed:   // есть изменения, перерисовать
    glutPostRedisplay();
    break;
  default:
    assert(false);
  }  
}


/// перезапуск (сортировать заново)
void restart()
{
  scene.resort( elements_to_be_sorted, algorithms.begin(), algorithms.end() );
  timer.restart();
}

/// обработать нажатие "обычной" клавиши
void on_key(unsigned char key, int x, int y)
{
  switch (key)
  {
  case '+': // ускорить демонстрацию в 1.5 раза
    timer.rate_proxy().multiply (1.5);
    break;

  case '-': // замедлить демонстрацию в 1.5 раза
    timer.rate_proxy().multiply (2./3.);
    break;

  case 27: case 'q': case 'Q': case 'й': case 'Й':
    glutLeaveMainLoop();
    break;

  case ' ':
    restart();
    break;

  case 'p': case 'P': case 'з': case 'З':
    scene.toggle_pause();
    break;

  case 't': case 'T': case 'е': case 'Е':
    scene.toggle_titles();
    glutPostRedisplay();
    break;

  // выбор генератора + рестарт
  case '1': case '2': case '3': case '4': case '5':
  case '6': case '7': case '8': case '9':
    scene.change_generator(key - '1');
    restart();
    break;

  case '0':
    scene.change_generator(10);
    restart();
    break;

  case '\r': case '\n':
    // прочитать:
    {
      using std::cin;
      using std::cout;
      // ... скорость,
      double speed = 20.;
      cout << "\nspeed = ";
      cin >> speed;
      if (cin)
      {
        speed = clamp(speed, 0.1, 1e+6);
        cout << " => " << speed << '\n';
        timer.rate_proxy() = Rate(speed);
      }
      else
      {
        cin.clear();
        cin.ignore();
      }

      // ... количество,
      int elms = 64;
      cout << "\nelements = ";
      cin >> elms;
      if (cin)
      {
        elements_to_be_sorted = clamp(elms, 4, 8192);
        cout << " => " << elements_to_be_sorted << '\n';
      }
      else
      {
        cin.clear();
        cin.ignore();
      }

      // ... список алгоритмов сортировок
      cout << "\nAlgorithms (delimit with newline, finish with semicolon):\n";
      algorithms.clear();
      while (true)
      {
        std::string name;
        std::getline (cin, name);
        if (!cin.good() || name == "end" || name == ";")
          break;
        algorithms.emplace_back(std::move(name));
      }
      cout << std::endl;
    }

    restart();
    break;

  default:
    return;
  }
}


//
// main
//

int main(int argc, char **argv)
{
  // базовая инициализация
  glutInit(&argc, argv);  // обработка параметров командной строки

  // создать окно с привязанным контекстом OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(800, 593);
  glutCreateWindow("Sort visualization");

  // параметры OpenGL для данного примера
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  
  // привязка функций-обработчиков событий
  glutDisplayFunc(repaint);
  glutKeyboardFunc(on_key);
  glutIdleFunc(frame_move);

  // старт
  std::cout.precision(3);
  std::cout << "Starting SortVis...\n";
  timer.rate_proxy().multiply (20.); // 20 событий в секунду
  scene.print_supported_sorts();
  restart();
  glutMainLoop();
  std::cout << "\nbye!";
  return 0;
}
