﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Game.hpp
 *
 * Модель игры. Реализация движения змейки.
 */
#pragma once
#include <algorithm>
#include <vector>
#include <utility>
#include <limits>

#include "ViewModelBase.hpp"


/// модель игрового поля, хранит массив ячеек Cell
/**
 * Массив ячеек организован в "матрицу", которую можно адресовать
 * индексами (x, y) с помощью оператора ().
 */
class Field
{
public:
  /// координаты на поле
  /** 
   * Работа с ними опирается на "оборачивающую" арифметику в дополнительном коде
   * (переполнения и преобразование отрицательных к беззнаковому типу).
   */
  using Position = std::pair<Index, Index>;

  /// задать размеры поля по горизонтали и вертикали
  /**
   * Заполняет поле пустыми ячейками.
   *
   * @param horz горизонтальный размер, индекс x от 0 до horz - 1
   * @param vert вертикальный размер, индекс y от 0 до vert - 1
   */
  Field(Index horz, Index vert)
    : xsize_(horz), ysize_(vert), cells(horz * vert, CELL_EMPTY)
  {}

  /// получить размер по горизонтали
  Index xsize() const
  {
    return xsize_;
  }

  /// получить размер по вертикали
  Index ysize() const
  {
    return ysize_;
  }

  /// получить клетку по координатам, бросает исключение для плохих x, y
  Cell operator()(Index x, Index y) const
  {
    return cells.at(x + xsize_ * y);
  }
  
  /// получить клетку по координатам, бросает исключение для плохих x, y
  Cell& operator()(Index x, Index y)
  {
    return cells.at(x + xsize_ * y);
  }

  /// получить клетку по координатам, бросает исключение для плохих координат
  Cell operator[](const Position &p) const
  {
    return (*this)(p.first, p.second);
  }

  /// получить клетку по координатам, бросает исключение для плохих координат
  Cell& operator[](const Position &p)
  {
    return (*this)(p.first, p.second);
  }
  

  /// пройтись по всем ячейкам поля, visitor принимает параметры (dx, dy, cell)
  /**
   * В начале перебора: dx == 0 и dy == 0;
   * затем передаются смещения в клетках от предыдущей переданной клетки.
   * Так сделано для упрощения использования функции glRasterPos.
   */
  template <class Visitor>
  void scan(Visitor visitor) const
  {
    visitor (0, 0, cells[0]);
    for (Index x = 0, i = 1, sz = cells.size(); i < sz; ++i)
    {
      if (++x == xsize_)
      {
        visitor(1 - xsize_, 1, cells[i]);
        x = 0;
      }
      else
      {
        visitor(1, 0, cells[i]);
      }
    }
  }

private:
  Index xsize_, ysize_;    // размеры поля
  std::vector<Cell> cells; // ячейки поля
};


/// состояние "змеи"
class Snake
{
public:
  /// позиция на поле
  using Position = Field::Position;

  /// по умолчанию двигается из левого нижнего угла (0, 0) вверх по y
  /**
   * @param field поле, на котором разместить змею
   * @param snake_len длина змеи, змея укладывается горизонтально
   */
  explicit Snake(Field &field, Index snake_len = 5)
    : field(field)
  {
    create_snake(snake_len);
  }

  /// сбросить змею
  /**
   * Змея удаляется с поля field, затем действует как конструктор.
   */
  void reset(Index snake_len = 0)
  {
    for (const auto &p : positions)
      field[p] = CELL_EMPTY;
    create_snake(snake_len);
  }

  /// двигаться "вверх"
  void up()
  {
    vx =  0;
    vy = +1;
    check_reverse();
  }

  /// двигаться "вниз"
  void down()
  {
    vx =  0;
    vy = -1;
    check_reverse();
  }
  
  /// двигаться "влево"
  void left()
  {
    vx = -1;
    vy =  0;
    check_reverse();
  }

  /// двигаться "вправо"
  void right()
  {
    vx = +1;
    vy =  0;
    check_reverse();
  }

  /// передвинуться на один шаг вперёд
  /**
   * @return пройденная клетка (если пустая -- прошла, если змея -- столкновение, кролик -- съел)
   */
  Cell step()
  {
    const Position new_head = moved(positions.front());
    if (field.xsize() <= new_head.first || field.ysize() <= new_head.second)
      return CELL_SNAKE;
    return warp_head(new_head);
  }

  /// телепортировать голову змеи
  /**
   * @return старое значение клетки, которую заняла голова змеи
   */
  Cell warp_head(const Position &new_head)
  {
    switch (const Cell next_cell = field[new_head])
    {
    case CELL_SNAKE:
      return CELL_SNAKE;

    case CELL_EMPTY:
      field[positions.back()] = CELL_EMPTY;
      positions.pop_back();
      // fall-through
    default:
      field[new_head] = CELL_SNAKE;
      positions.insert(positions.begin(), new_head);
      return next_cell;
    }
  }

  /// get snake length
  Index length() const { return positions.size(); }

private:

  using Positions = std::vector<Position>;
  /// позиции клеток, занимаемых змеёй, голова на индексе 0
  Positions positions;
  Index vx, vy; // скорость по x, по y
  Field &field; // привязка к полю

  /// создать змею
  void create_snake(Index snake_len)
  {
    vx = 0;
    vy = 1;
    positions.resize(snake_len);
    for (Index i = 0; i < snake_len; ++i)
    {
      const Position block(snake_len - i - 1, 0);
      positions[i] = block;
      field[block] = CELL_SNAKE;
    }
  }

  /// сдвинутая на vx, vy позиция
  Position moved(const Position &p) const
  {
    return Position(p.first + vx, p.second + vy);
  }

  /// обратить порядок, если нужно
  void check_reverse()
  {
    if (1 < positions.size() && moved(positions.front()) == positions[1])
    {
      std::reverse(positions.begin(), positions.end());
      // даже после обращения порядка, "голова" может двигаться в "шею"
      if (moved(positions.front()) == positions[1])
        vx = positions.front().first  - positions[1].first ,
        vy = positions.front().second - positions[1].second;
    }
  }
};


/// описание уровня
struct Level
{
  Index start_length, ///< начальная длина змеи
        max_length;   ///< максимальная длина змеи, число кроликов == max_length - start_length

  double speed;       ///< скорость в ячейках в секунду
  
  Score max_movement, ///< максимальное количество ячеек, засчитываемых в очки
        penalty;      ///< штраф за столкновение (домножается на номер уровня)

  Level()
    : start_length(4),
      max_length(10),
      speed(4.0),
      max_movement(1200),
      penalty(300)
  {}
};


/// класс представляющий модель игры "Змейка"
class Zmeika_game
{
public:
  Zmeika_game();

  /// собственно игровое поле
  Field field;

  /// узнать текущий уровень
  Index get_level() const { return level; }
  /// узнать текущее количество очков
  Score get_score() const { return score; }
  /// узнать максимальное количество очков игрока за всю игру (очки могут сниматься за столкновения)
  Score get_score_max() const { return score_max; }

  /// змея будет двигаться вверх
  void up()    { snake.up();    }
  /// змея будет двигаться вниз
  void down()  { snake.down();  }
  /// змея будет двигаться влево
  void left()  { snake.left();  }
  /// змея будет двигаться вправо
  void right() { snake.right(); }

  /// начать новую игру
  void start()
  {
    read_levels();
    new_rabbit(CELL_RABBIT);
    select_level(1);
    score = score_max = 0;
  }

  /// телепортировать голову змеи
  /**
   * @return true если произошло столкновение (останов змеи)
   */
  bool warp_snake(const Field::Position &new_head)
  {
    return moved_into(snake.warp_head(new_head));
  }

protected:

  /// узнать временнОй шаг (в секундах) текущего уровня
  double get_time_step() const { return time_step; }
  /// выполнить шаг
  bool do_step()
  {
    return moved_into(snake.step());
  }

private:

  /// прочитать описания уровней
  void read_levels();

  /// поставить нового кролика
  void new_rabbit(Cell rabbit_type);
  /// выбрать уровень
  void select_level(Index next_level);

  /// реакция змеи на движение в клетку с заданным содержимым
  /**
   * @return true если произошло столкновение (останов змеи)
   */
  bool moved_into(const Cell);

  /// добавить очки
  void add_score(Score amount);
  /// снять очки
  void sub_score(Score amount);

  /// текущий уровень
  Index level;
  /// текущие очки игрока
  Score score;
  /// максимальные очки игрока за игру
  Score score_max;

  /// премиальные за уровень
  Score level_score;
  /// очки за движение в ту или иную клетку (по её коду Cell)
  Score move_score[CELL_TYPES];
  /// текущая скорость змеи определяется задержкой между кадрами (секунды)
  double time_step;
  /// набор уровней
  std::vector<Level> levels;
  /// пройденные шаги
  Index steps;
  /// состояние змеи
  Snake snake;
  // Осторожно: конструктор принимает field, поэтому snake дб расположена
  // в классе ниже, чем field, чтобы конструктор field отработал до конструктора snake.
};
