﻿/*
 * This file is part of AV Orchid.
 * Copyright (c) 2013, Dmitri R. Kuvshinov <evetro.here@gmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 * @file Screen.cpp
 *
 * Реализация элементов Screen.hpp.
 */
#include <GL/freeglut.h>
#include "Screen.hpp"

#include <algorithm>
#include <mutex>


//
// рисование текста

/// определить количество строк в тексте (сколько строчек по вертикали будет занимать изображение)
inline std::size_t lines_in_text(const std::string &text)
{
  return 1 + std::count(text.begin(), text.end(), '\n');
}

// вывести текст в заданный прямоугольник (внутри уже установленной системы координат)
void draw_stroke_string(void *font, const std::string &text,
                      float left, float bottom, float right, float top)
{
  // функции, вычисляющие "размеры" (в пикселях), не задействуют текущую матрицу,
  // а просто берут значения из зашитых внутри FreeGlut массивов
  auto cstr = reinterpret_cast<const unsigned char*>(text.c_str());
  const GLfloat wdth = static_cast<GLfloat>(glutStrokeLength(font, cstr)), // ширина области рисования
                hght = glutStrokeHeight(font) * lines_in_text(text);       // высота области рисования

  // теперь желательно вписать эти размеры в заданный прямоугольник
  glPushMatrix();
  glTranslatef(left, bottom, 0.0f);
  const GLfloat scale_x = left < right? (right - left) / wdth : 1.0f,
                scale_y = bottom < top? (top - bottom) / hght : 1.0f;

  if (scale_x != 1.0f || scale_y != 1.0f)
    glScalef(scale_x, scale_y, 1.0f);

  // отрисовка текста
  glutStrokeString(font, cstr);
  // восстановление текущей матрицы
  glPopMatrix();
}


//
// подгонка размеров окна

void fit_window(float target_width, float target_height)
{
  const float scr_width   = float(Drawing_area_info::screen_width()),
              scr_height  = Drawing_area_info::screen_height() * 0.92f; 
  // магическое число-поправка, рассчитана на горизонтальную панель задач
  // наверняка где-нибудь будет мешать...

  if (scr_width < target_width)
    target_height *= scr_width / target_width, 
    target_width = scr_width;
  if (scr_height < target_height)
    target_width *= scr_height / target_height,
    target_height = scr_height;

  glutPositionWindow(static_cast<int>(0.5f * (scr_width - target_width)),
                     static_cast<int>(0.5f * (scr_height - target_height)));
  glutReshapeWindow(static_cast<int>(target_width), static_cast<int>(target_height));
  // возможно, нужно вводить поправку на ширину границ
}


//
// получение размеров области рисования окна, экрана (в пикселях)

int Drawing_area_info::x()
{
  return glutGet(GLUT_WINDOW_X);
}

int Drawing_area_info::y()
{
  return glutGet(GLUT_WINDOW_Y);
}

int Drawing_area_info::width()
{
  return glutGet(GLUT_WINDOW_WIDTH);
}

int Drawing_area_info::height()
{
  return glutGet(GLUT_WINDOW_HEIGHT);
}

int Drawing_area_info::border_width()
{
  return glutGet(GLUT_WINDOW_BORDER_WIDTH);
}

int Drawing_area_info::border_height()
{
  return glutGet(GLUT_WINDOW_BORDER_HEIGHT);
}

int Drawing_area_info::screen_width()
{
  return glutGet(GLUT_SCREEN_WIDTH);
}

int Drawing_area_info::screen_height()
{
  return glutGet(GLUT_SCREEN_HEIGHT);
}


//
// AScreen

namespace
{

// текущий экран -- эта глобальная переменная и есть причина невозможности
// переключение экранов из разных параллельно исполняющихся потоков
AScreen *curScreen;
std::mutex curScreen_mut; // на случай попыток переключения экрана из разных нитей одновременно

}

using Lock = std::lock_guard<std::mutex>;

bool AScreen::has_active()
{
  Lock lm(curScreen_mut);
  return curScreen != nullptr;
}

AScreen& AScreen::get_active()
{
  Lock lm(curScreen_mut);
  return *curScreen;
}


AScreen::~AScreen()
{
  Lock lm(curScreen_mut);
  if (curScreen == this)
    current_screen_select(nullptr);
}


void AScreen::select()
{
  Lock lm(curScreen_mut);
  if (curScreen != this)
  {
    // отключить
    if (curScreen)         
      curScreen->on_deselect();

    current_screen_select(this);
  }
}


void AScreen::current_screen_select(AScreen *newScreen)
{
  if (newScreen == curScreen)
    return;

  if (!newScreen) // => curScreen != nullptr
  {
    // отвязать обработчики
    glutIdleFunc          (nullptr);
    glutDisplayFunc       (nullptr);
    glutReshapeFunc       (nullptr);
    glutVisibilityFunc    (nullptr);
    glutKeyboardFunc      (nullptr);
    glutKeyboardUpFunc    (nullptr);
    glutSpecialFunc       (nullptr);
    glutSpecialUpFunc     (nullptr);
    glutEntryFunc         (nullptr);
    glutPassiveMotionFunc (nullptr);
    glutMotionFunc        (nullptr);
    glutMouseFunc         (nullptr);
    glutMouseWheelFunc    (nullptr);
  }
  else if (!curScreen) // => newScreen != nullptr
  {
    // привязать обработчики
    glutIdleFunc          (&AScreen::on_idle_wrapper);
    glutDisplayFunc       (&AScreen::on_display_wrapper);
    glutReshapeFunc       (&AScreen::on_reshape_wrapper);
    glutVisibilityFunc    (&AScreen::on_visibility_wrapper);
    glutKeyboardFunc      (&AScreen::on_char_wrapper);
    glutKeyboardUpFunc    (&AScreen::on_char_up_wrapper);
    glutSpecialFunc       (&AScreen::on_special_key_wrapper);
    glutSpecialUpFunc     (&AScreen::on_special_key_up_wrapper);
    glutEntryFunc         (&AScreen::on_entry_wrapper);
    glutPassiveMotionFunc (&AScreen::on_hover_wrapper);
    glutMotionFunc        (&AScreen::on_drag_wrapper);
    glutMouseFunc         (&AScreen::on_mouse_button_wrapper);
    glutMouseWheelFunc    (&AScreen::on_mouse_wheel_wrapper);
  }

  curScreen = newScreen;
  // подключить
  if (curScreen)
  {
    curScreen->on_select();
    glutShowWindow();
  }
}

// готовые обработчики

void AScreen::on_reshape(int width, int height)
{
  glViewport(0, 0, width, height);
}

void AScreen::on_display_before()
{
  glClear(GL_COLOR_BUFFER_BIT); // сброс буфера цвета
}

void AScreen::on_display_after()
{
  glutSwapBuffers(); // обмен буфера
}

// обёртки обработчиков событий -- в основном занимаются простой переадресацией

void AScreen::on_idle_wrapper()
{
  curScreen->on_idle();
}

void AScreen::on_display_wrapper()
{
  curScreen->on_display_before();
  curScreen->on_display();
  glFinish();
  curScreen->on_display_after();
}

void AScreen::on_reshape_wrapper(int width, int height)
{
  curScreen->on_reshape(width, height);
}

void AScreen::on_position_wrapper(int x, int y)
{
  curScreen->on_position(x, y);
}

void AScreen::on_visibility_wrapper(int status)
{
  curScreen->on_visibility(status == GLUT_VISIBLE);
}

void AScreen::on_char_wrapper(unsigned char code, int mouse_x, int mouse_y)
{
  curScreen->on_char(code, mouse_x, mouse_y);
}

void AScreen::on_char_up_wrapper(unsigned char code, int mouse_x, int mouse_y)
{
  curScreen->on_char_up(code, mouse_x, mouse_y);
}

void AScreen::on_special_key_wrapper(int code, int mouse_x, int mouse_y)
{
  curScreen->on_special_key(code, mouse_x, mouse_y);
}

void AScreen::on_special_key_up_wrapper(int code, int mouse_x, int mouse_y)
{
  curScreen->on_special_key_up(code, mouse_x, mouse_y);
}

void AScreen::on_entry_wrapper(int status)
{
  curScreen->on_entry(status == GLUT_ENTERED);
}

void AScreen::on_hover_wrapper(int mouse_x, int mouse_y)
{
  curScreen->on_hover(mouse_x, mouse_y);
}

void AScreen::on_drag_wrapper(int mouse_x, int mouse_y)
{
  curScreen->on_drag(mouse_x, mouse_y);
}

void AScreen::on_mouse_button_wrapper(int button, int state, int x, int y)
{
  if (state == GLUT_DOWN)
    curScreen->on_mouse(button, x, y);
  else
    curScreen->on_mouse_up(button, x, y);
}

void AScreen::on_mouse_wheel_wrapper(int wheel, int direction, int x, int y)
{
  curScreen->on_mouse_wheel(wheel, direction, x, y);
}
